#include <asm/semaphore.h>
#include <asm/uaccess.h>   /* copy_from_user, copy_to_user */
#include <linux/errno.h>
#include <linux/fs.h>      /* inode, file, file_operations */
#include <linux/kernel.h>  /* printk() */
#include <linux/module.h> 
#include <linux/poll.h>    /* POLLIN ... */
#include <linux/wait.h>    /* wait_queue_head_t */
#include <linux/sched.h>   /* wait_event_interruptible, wake_up_interruptible */


MODULE_AUTHOR( "Ryo" );
MODULE_DESCRIPTION( "Sample character device driver." );
MODULE_LICENSE( "GPL" );


static int   devmajor = 77;
static char* devname  = "chardev";
static char* msg      = "module [poll.o]";


#define MAXBUF 32
static unsigned char devbuf[ MAXBUF ];
static int buf_pos;

static struct semaphore chardev_sem;
static wait_queue_head_t write_q;
static wait_queue_head_t read_q;


/*
 * ϥեޥåѴؿ
 */
static void 
log_format( char* body )
{
	if ( current != NULL ) {
		printk( KERN_INFO "%s : PID = %-5i cmdline = \"%-10s\" : %s\n", 
			msg, current->pid, current->comm, body );
	}
	else {
		printk( KERN_INFO "%s : %s\n", msg, body );
	}
}


/*
 * open()
 */
static int 
chardev_open( struct inode* inode, struct file* filp )
{
	log_format( "open()  called" );
	return 0;
}


/*
 * release()
 */
static int 
chardev_release( struct inode* inode, struct file* filp )
{
	log_format( "close() called" );
	return 0;
}


/*
 * read()
 * 桼ץФХåեƤž롣
 * žƤХåեä롣
 */
static ssize_t 
chardev_read( struct file* filp, char* buf, size_t count, loff_t* pos )
{
	int copy_len;
	int i;

	log_format( "read()  called" );

	if ( down_interruptible( &chardev_sem ) ) {
		log_format( "down_interruptible for read failed" );
		return -ERESTARTSYS;
	}

	while ( buf_pos == 0 ) {
		up( &chardev_sem );

		if ( filp->f_flags & O_NONBLOCK ) {
			log_format( "read return because O_NONBLOCK is specified" );
			return -EAGAIN;
		}

		log_format( "read process going to sleep" );

		if ( wait_event_interruptible( read_q, ( buf_pos != 0 ) ) )
			return -ERESTARTSYS;

		if ( down_interruptible( &chardev_sem ) ) {
			log_format( "down_interruptible for read failed" );
			return -ERESTARTSYS;
		}
	}


	log_format( "read process going to read" );

	if ( count > buf_pos )
		copy_len = buf_pos;
	else
		copy_len = count;

	if ( copy_to_user( buf, devbuf, copy_len ) ) {
		up( &chardev_sem );
		log_format( "copy_to_user failed" );
		return -EFAULT;
	}

	*pos += copy_len;

	for ( i = copy_len; i < buf_pos; i ++ )
		devbuf[ i - copy_len ] = devbuf[i];

	buf_pos -= copy_len;

	up( &chardev_sem );
	wake_up_interruptible( &write_q );

	return copy_len;
}


/*
 * write()
 * 桼ץž줿ƤХåե˽񤭹ࡣ
 */
static ssize_t 
chardev_write( struct file* filp, const char* buf, size_t count, loff_t* pos )
{
	int copy_len;

	log_format( "write() called" );

	if ( down_interruptible( &chardev_sem ) ) {
		log_format( "down_interruptible for write failed" );
		return -ERESTARTSYS;
	}

	while ( buf_pos == MAXBUF ) {
		up( &chardev_sem );

		if ( filp->f_flags & O_NONBLOCK ) {
			log_format( "write return because O_NONBLOCK is specified" );
			return -EAGAIN;
		}

		log_format( "write process going to sleep" );

		if ( wait_event_interruptible( write_q, ( buf_pos != MAXBUF ) ) )
			return -ERESTARTSYS;

		if ( down_interruptible( &chardev_sem ) ) {
			log_format( "down_interruptible for write failed" );
			return -ERESTARTSYS;
		}
	}


	log_format( "write process going to write" );

	if ( count > ( MAXBUF - buf_pos ) )
		copy_len = MAXBUF - buf_pos;
	else
		copy_len = count;

	if ( copy_from_user( devbuf + buf_pos, buf, copy_len ) ) {
		up( &chardev_sem );
		log_format( "copy_from_user failed" );
		return -EFAULT;
	}

	*pos    += copy_len;
	buf_pos += copy_len;

	up( &chardev_sem );
	wake_up_interruptible( &read_q );

	return copy_len;
}


/*
 * poll()
 * select / poll ƥॳȯԻ˼¹Ԥ롣
 */
static unsigned int 
chardev_poll( struct file* filp, poll_table* wait )
{
	unsigned int retmask = 0;

	log_format( "poll()  called" );

	poll_wait( filp, &read_q,  wait );
	poll_wait( filp, &write_q, wait );

	if ( buf_pos != 0 ) {
		retmask |= ( POLLIN  | POLLRDNORM );
		log_format( "POLLIN  | POLLRDNORM" );
	}

	if ( buf_pos != MAXBUF ) {
		retmask |= ( POLLOUT | POLLWRNORM );
		log_format( "POLLOUT | POLLWRNORM" );
	}

	/* 
	   ɤ߹ߥǥХ EOF ξ retmask  POLLHUP 
	   ǥХ顼֤Ǥ POLLERR 
	   out-of-band ǡɤ߽Ф POLLPRI 
	*/

	return retmask;
}


static struct file_operations chardev_fops = 
{
	owner   : THIS_MODULE,
	read    : chardev_read,
	write   : chardev_write,
	poll    : chardev_poll,
	open    : chardev_open,
	release : chardev_release,
};


/*
 * ⥸塼ν
 * insmod ˸ƤФ
 */
int 
init_module( void )
{
	if ( register_chrdev( devmajor, devname, &chardev_fops ) ) {
		log_format( "register_chrdev failed" );
		return -EBUSY;
	}

	sema_init( &chardev_sem, 1 );

	init_waitqueue_head( &write_q );
	init_waitqueue_head( &read_q );

	log_format( "loaded  into kernel" );
	return 0;
}


/*
 * ⥸塼β
 * ⥸塼λȿ 0 ǤСrmmod ˸ƤФ
 */
void 
cleanup_module( void )
{
	if ( unregister_chrdev( devmajor, devname ) ) {
		log_format( "unregister_chrdev failed" );
		/* к̵ ... */
	}

	log_format( "removed from kernel" );
}


/* End of poll.c */
