#include <asm/io.h>          /* outb(), inb() */
#include <asm/uaccess.h>     /* copy_from_user, copy_to_user */
#include <linux/fs.h>        /* inode, file, file_operations, ... */
#include <linux/ioport.h>    /* check_region(), request_region(), release_region() */
#include <linux/kernel.h>    /* printk() */
#include <linux/module.h>
#include <linux/sched.h>     /* task_struct */
#include <linux/timer.h> 

#include "irq_io.h"


MODULE_LICENSE( "GPL" );


static int   devmajor = 77;
static char* devname  = "chardev";
static char* msg      = "module [irq.o]";


/* 100 tick = 1 sec. */
static int timeout = 100; 
MODULE_PARM( timeout, "i" );

static struct timer_list chardev_timer;
static int timerstop = 0;

static unsigned int initial_data;
static unsigned int initial_control;

static void chardev_timer_handler( unsigned long data );
static void chardev_del_timer();


/*
 * ϥեޥåѴؿ
 */
static void 
log_format( char* body )
{
	if ( current != NULL ) {
		printk( KERN_INFO "%s : PID = %-5i cmdline = \"%-10s\" : %s\n", 
			msg, current->pid, current->comm, body );
	}
	else {
		printk( KERN_INFO "%s : %s\n", msg, body );
	}
}


/*
 * ߥϥɥ
 */
static void
chardev_interrupt( int irq, void* dev_id, struct pt_regs* regs )
{
	io_write_xor( IOPORT_DATA, ( 1 << DATA_LED_3 ) );
}


/* 
 * ޡϿ 
 */
static void 
chardev_timer_register( void )
{
	init_timer( &chardev_timer );

	chardev_timer.expires  = jiffies + timeout;
	chardev_timer.data     = ( unsigned long )jiffies;
	chardev_timer.function = chardev_timer_handler;

	add_timer( &chardev_timer );
}


/*
 * ޡؿ
 * ߥե饰ꤷdel_timer_sync Ǵ¸Υޡؿ
 * ʼ¹ξϽλޤԤġˡ
 */
static void
chardev_del_timer()
{
	timerstop = 1;
	del_timer_sync( &chardev_timer );
}


/* 
 * ॢȽ 
 */
static void 
chardev_timer_handler( unsigned long data )
{
	io_write_xor( IOPORT_DATA, ( 1 << DATA_LED_2 ) );

	if ( timerstop ) {
		printk( KERN_INFO "%s : Timer is not to be restarted.\n", msg );
		return;
	}

	chardev_timer_register();
}


/*
 * open()
 */
static int 
chardev_open( struct inode* inode, struct file* filp )
{
	log_format( "open()  called" );
	return 0;
}


/*
 * release()
 */
static int 
chardev_release( struct inode* inode, struct file* filp )
{
	log_format( "close() called" );
	return 0;
}


/*
 * read()
 */
static ssize_t 
chardev_read( struct file* filp, char* buf, size_t count, loff_t* pos )
{
	char status[32];
	int copy_len = 0;
	int len = 0;

	log_format( "read()  called" );

	if ( *pos == 0 ) {
		if ( io_is_button_pressed() )
			len = sprintf( status, "[Switch] pressed.\n" );
		else
			len = sprintf( status, "[Switch] NOT pressed.\n" );
	}

	copy_len = ( len > count ) ? count : len;

	if ( copy_to_user( buf, status, copy_len ) ) {
		log_format( "copy_to_user failed\n" );
		return -EFAULT;
	}

	*pos += copy_len;

	return copy_len;
}


/*
 * write()
 */
static ssize_t 
chardev_write( struct file* filp, const char* buf, size_t count, loff_t* pos )
{
	log_format( "write() called" );
	io_write_xor( IOPORT_DATA, ( 1 << DATA_LED_1 ) );
	return count;
}


static struct file_operations chardev_fops = 
{
	owner   : THIS_MODULE,
	read    : chardev_read,
	write   : chardev_write,
	open    : chardev_open,
	release : chardev_release,
};


/*
 * ⥸塼ν
 * insmod ˸ƤФ
 */
int 
init_module( void )
{
	EXPORT_NO_SYMBOLS;

	if ( register_chrdev( devmajor, devname, &chardev_fops ) ) {
		log_format( "register_chrdev failed" );
		return -EBUSY;
	}


	if ( request_region( IOPORT_BASE, IOPORT_RANGE, devname ) == NULL ) {
		log_format( "request_region returned NULL" );
		unregister_chrdev( devmajor, devname );
		return -EBUSY;
	}


	if ( request_irq( IRQ_NUMBER, chardev_interrupt, 0, devname, NULL ) ) {
		log_format( "request_irq failed" );
		release_region( IOPORT_BASE, IOPORT_RANGE );
		unregister_chrdev( devmajor, devname );
		return -EBUSY;
	}

	initial_data    = inb( IOPORT_DATA );
	initial_control = inb( IOPORT_CONTROL );

	io_write_or( IOPORT_DATA, ( 1 << DATA_PULLUP ) );
	io_enable_irq();

	chardev_timer_register();

	log_format( "loaded  into kernel" );
	return 0;
}


/*
 * ⥸塼β
 * ⥸塼λȿ 0 ǤСrmmod ˸ƤФ
 */
void 
cleanup_module( void )
{
	chardev_del_timer();
	free_irq( IRQ_NUMBER, NULL );

	io_write_set( IOPORT_CONTROL, initial_control );
	io_write_set( IOPORT_DATA,    initial_data );

	release_region( IOPORT_BASE, IOPORT_RANGE );

	if ( unregister_chrdev( devmajor, devname ) ) {
		log_format( "unregister_chrdev failed" );
		/* к̵ ... */
	}

	log_format( "removed from kernel" );
}


/* End of irq_main.c */
